"""Command-line interface for pipes."""

import argparse
import curses
import sys
from pipes import __version__
from pipes.config import load_config, save_config
from pipes.pipes import PipesScreen
from typing import Any


def parse_args() -> argparse.Namespace:
    """Parse command-line arguments.

    Returns:
        Parsed arguments
    """
    parser = argparse.ArgumentParser(
        description="Basically pipes.sh but rewritten in Python",
        formatter_class=argparse.HelpFormatter,
    )

    parser.add_argument("-p", "--pipes", type=int, help="number of pipes")
    parser.add_argument("-f", "--fps", type=int, help="frames per second (20-100)")
    parser.add_argument("-s", "--steady", type=int, help="steadiness (5-15)")
    parser.add_argument("-r", "--limit", type=int, help="character limit before reset")
    parser.add_argument("-R", "--random", action="store_true", help="random start")
    parser.add_argument("-B", "--no-bold", action="store_true", help="disable bold")
    parser.add_argument("-C", "--no-color", action="store_true", help="disable color")
    parser.add_argument(
        "-P",
        "--pipe-style",
        type=int,
        choices=range(10),
        help="change pipe style (0-9)",
    )
    parser.add_argument(
        "-K", "--keep-style", action="store_true", help="keep style on wrap"
    )
    parser.add_argument(
        "-S",
        "--save-config",
        action="store_true",
        help="save current settings as default",
    )
    parser.add_argument(
        "-v", "--version", action="version", version=f"pipes-py v{__version__}"
    )

    return parser.parse_args()


def main() -> None:
    """Main entry point for the application."""
    args = parse_args()
    config = load_config()

    # Update config with command line arguments
    if args.pipes is not None:
        config.pipes = max(1, args.pipes)
    if args.fps is not None:
        config.fps = max(20, min(100, args.fps))
    if args.steady is not None:
        config.steady = max(5, min(15, args.steady))
    if args.limit is not None:
        config.limit = max(0, args.limit)
    if args.random:
        config.random_start = True
    if args.no_bold:
        config.bold = False
    if args.no_color:
        config.color = False
    if args.keep_style:
        config.keep_style = True
    if args.pipe_style is not None:
        config.pipe_types = [args.pipe_style]

    if args.save_config:
        save_config(config)

    try:
        curses.wrapper(lambda screen: run_pipes(screen, config))
    except KeyboardInterrupt:
        sys.exit(0)


def run_pipes(screen: Any, config: Any) -> None:
    """Run the pipes animation.

    Args:
        screen: Curses screen object
        config: Pipe configuration
    """
    pipes = PipesScreen(screen, config)
    while pipes.update():
        pass


if __name__ == "__main__":
    main()



"""Core pipe state management."""

import random
import time
from pipes.renderer import Renderer
from pipes.types import Direction, Pipe, PipeConfig
from typing import Any


class PipesScreen:
    """Manages the state and behavior of pipes on the screen."""

    def __init__(self, screen: Any, config: PipeConfig) -> None:
        """Initialize the pipes screen.

        Args:
            screen: Curses screen object
            config: Pipe configuration
        """
        self.screen = screen
        self.config = config
        self.renderer = Renderer(screen, config)
        self.pipes: list[Pipe] = []

        self.height, self.width = screen.getmaxyx()
        self.count = 0
        self.delay = 1.0 / self.config.fps

        self._init_pipes()

    def _init_pipes(self) -> None:
        """Initialize all pipes with random or centered positions."""
        h, w = self.screen.getmaxyx()
        for _ in range(self.config.pipes):
            direction = (
                Direction(random.randrange(4))
                if self.config.random_start
                else Direction.UP
            )
            x = random.randrange(w) if self.config.random_start else w // 2
            y = random.randrange(h) if self.config.random_start else h // 2

            pipe_type = random.choice(self.config.pipe_types)
            color = random.choice(self.config.colors)

            self.pipes.append(
                Pipe(
                    x=x,
                    y=y,
                    direction=direction,
                    pipe_type=pipe_type,
                    color=color,
                    attr=self.renderer.get_color_attr(color),
                )
            )

    def update(self) -> bool:
        """Update pipes state and render.

        Returns:
            False if the user wants to quit, True otherwise
        """
        key = self.screen.getch()
        if key != -1 and not self._handle_key(key):
            return False

        new_h, new_w = self.screen.getmaxyx()
        if new_h != self.height or new_w != self.width:
            self.height, self.width = new_h, new_w
            self.renderer.clear()

        self._update_pipes()
        self.renderer.refresh()

        self.count += len(self.pipes)
        if self.config.limit > 0 and self.count >= self.config.limit:
            self.renderer.clear()
            self.count = 0

        time.sleep(self.delay)
        return True

    def _update_pipes(self) -> None:
        """Update position and direction of all pipes."""
        for pipe in self.pipes:
            x, y = pipe.x, pipe.y
            old_direction = pipe.direction

            # Update position based on direction
            if old_direction % 2:  # RIGHT or LEFT
                x += -old_direction + 2
            else:  # UP or DOWN
                y += old_direction - 1

            # Handle wrapping
            if x < 0 or x >= self.width or y < 0 or y >= self.height:
                if not self.config.keep_style:
                    pipe.pipe_type = random.choice(self.config.pipe_types)
                    pipe.color = random.choice(self.config.colors)
                    pipe.attr = self.renderer.get_color_attr(pipe.color)
                x %= self.width
                y %= self.height

            # Calculate new direction
            new_direction = old_direction
            if random.randrange(self.config.steady) <= 1:
                turn = 2 * random.randrange(2) - 1  # -1 or 1
                new_direction = Direction((old_direction + turn) % 4)

            # Draw pipe segment
            self.renderer.draw_pipe(pipe, old_direction, new_direction)

            # Update pipe state
            pipe.x = x
            pipe.y = y
            pipe.direction = new_direction

    def _update_pipe_colors(self) -> None:
        """Update all pipe colors after color/bold changes."""
        for pipe in self.pipes:
            pipe.attr = self.renderer.get_color_attr(pipe.color)

    def _handle_key(self, key: int) -> bool:
        """Handle keyboard input.

        Args:
            key: Key code from curses

        Returns:
            False if the user wants to quit, True otherwise
        """
        key_char = chr(key).upper() if 0 <= key <= 255 else ""

        if key_char == "P" and self.config.steady < 15:
            self.config.steady += 1
        elif key_char == "O" and self.config.steady > 3:
            self.config.steady -= 1
        elif key_char == "F" and self.config.fps < 100:
            self.config.fps += 1
            self.delay = 1.0 / self.config.fps
        elif key_char == "D" and self.config.fps > 20:
            self.config.fps -= 1
            self.delay = 1.0 / self.config.fps
        elif key_char == "B":
            self.config.bold = not self.config.bold
            self.renderer.reinit_colors()
            self._update_pipe_colors()
        elif key_char == "C":
            self.config.color = not self.config.color
            self.renderer.reinit_colors()
            self._update_pipe_colors()
        elif key_char == "K":
            self.config.keep_style = not self.config.keep_style
        elif key_char == "?" or key == 27:  # ESC
            return False
        return True



"""Terminal rendering logic for pipes."""

import contextlib
import curses
import locale
from pipes.types import Direction, Pipe, PipeConfig
from typing import Any

# Ensure correct handling of Unicode characters
locale.setlocale(locale.LC_ALL, "")

PIPE_SETS = [
    "┃┏ ┓┛━┓  ┗┃┛┗ ┏━",
    "│╭ ╮╯─╮  ╰│╯╰ ╭─",
    "│┌ ┐┘─┐  └│┘└ ┌─",
    "║╔ ╗╝═╗  ╚║╝╚ ╔═",
    "|+ ++-+  +|++ +-",
    "|/ \\ /-\\  \\|/\\ /-",
    ".o ....  .... .o",
    ".o oo.o  o.oo o.",
    "-\\ /\\|/  /-\\/ \\|",
    "╿┍ ┑┚╼┒  ┕╽┙┖ ┎╾",
]


class Renderer:
    """Handles terminal rendering for pipes."""

    def __init__(self, screen: Any, config: PipeConfig) -> None:
        """Initialize the renderer.

        Args:
            screen: Curses screen object
            config: Pipe configuration
        """
        self.screen = screen
        self.config = config
        self.color_pairs: dict[int, int] = {}
        self.sets = self._prepare_sets()

        # Screen setup
        curses.curs_set(0)
        screen.nodelay(True)
        screen.clear()

        self._init_colors()

    def _prepare_sets(self) -> list[str]:
        """Prepare pipe character sets.

        Returns:
            List of pipe characters
        """
        sets: list[str] = []
        for pipe_set in PIPE_SETS:
            sets.extend((pipe_set + " " * 16)[:16])
        return sets

    def _init_colors(self) -> None:
        """Initialize color pairs for rendering."""
        if not self.config.color or not curses.has_colors():
            self.color_pairs = dict.fromkeys(self.config.colors, curses.A_NORMAL)
            return

        curses.start_color()
        curses.use_default_colors()
        max_colors = min(curses.COLORS, 8)

        for idx, color in enumerate(self.config.colors):
            curses_color = color % max_colors
            pair_number = idx + 1
            curses.init_pair(pair_number, curses_color, -1)
            attr = curses.color_pair(pair_number)
            if self.config.bold:
                attr |= curses.A_BOLD
            self.color_pairs[color] = attr

    def get_color_attr(self, color: int) -> int:
        """Get the color attribute for a given color.

        Args:
            color: Color index

        Returns:
            Curses attribute value
        """
        return self.color_pairs[color]

    def draw_pipe(
        self, pipe: Pipe, old_direction: Direction, new_direction: Direction
    ) -> None:
        """Draw a pipe segment at the current position.

        Args:
            pipe: Pipe to draw
            old_direction: Previous direction
            new_direction: New direction
        """
        base = pipe.pipe_type * 16
        index = base + old_direction * 4 + new_direction
        char = self.sets[index] if index < len(self.sets) else "?"

        with contextlib.suppress(curses.error):
            self.screen.addstr(pipe.y, pipe.x, char, pipe.attr)

    def clear(self) -> None:
        """Clear the screen."""
        self.screen.clear()

    def refresh(self) -> None:
        """Refresh the screen."""
        self.screen.refresh()

    def reinit_colors(self) -> None:
        """Reinitialize colors (used when toggling color/bold)."""
        self._init_colors()



"""Configuration management for pipes."""

import json
import sys
from pathlib import Path
from pipes.types import PipeConfig


def get_config_dir() -> Path:
    """Get platform-specific configuration directory."""
    if sys.platform == "win32":
        local_app_data = Path.home() / "AppData" / "Local"
        return local_app_data / "pipes-py"
    return Path.home() / ".config" / "pipes-py"


CONFIG_DIR = get_config_dir()
CONFIG_FILE = CONFIG_DIR / "config.json"

DEFAULT_CONFIG = PipeConfig(
    pipes=1,
    fps=75,
    steady=13,
    limit=2000,
    random_start=False,
    bold=True,
    color=True,
    keep_style=False,
    colors=[1, 2, 3, 4, 5, 6, 7, 0],
    pipe_types=[0],
)


def load_config() -> PipeConfig:
    """Load configuration from file or return defaults."""
    if not CONFIG_FILE.exists():
        return DEFAULT_CONFIG

    try:
        with CONFIG_FILE.open("r") as f:
            data = json.load(f)
            return PipeConfig(
                pipes=data.get("pipes", DEFAULT_CONFIG.pipes),
                fps=data.get("fps", DEFAULT_CONFIG.fps),
                steady=data.get("steady", DEFAULT_CONFIG.steady),
                limit=data.get("limit", DEFAULT_CONFIG.limit),
                random_start=data.get("random_start", DEFAULT_CONFIG.random_start),
                bold=data.get("bold", DEFAULT_CONFIG.bold),
                color=data.get("color", DEFAULT_CONFIG.color),
                keep_style=data.get("keep_style", DEFAULT_CONFIG.keep_style),
                colors=data.get("colors", DEFAULT_CONFIG.colors),
                pipe_types=data.get("pipe_types", DEFAULT_CONFIG.pipe_types),
            )
    except (json.JSONDecodeError, OSError):
        return DEFAULT_CONFIG


def save_config(config: PipeConfig) -> None:
    """Save configuration to file."""
    CONFIG_DIR.mkdir(parents=True, exist_ok=True)
    try:
        data = {
            "pipes": config.pipes,
            "fps": config.fps,
            "steady": config.steady,
            "limit": config.limit,
            "random_start": config.random_start,
            "bold": config.bold,
            "color": config.color,
            "keep_style": config.keep_style,
            "colors": config.colors,
            "pipe_types": config.pipe_types,
        }
        with CONFIG_FILE.open("w") as f:
            json.dump(data, f, indent=2)
    except OSError:
        pass



"""Type definitions for the pipes application."""

from dataclasses import dataclass
from enum import IntEnum


class Direction(IntEnum):
    """Pipe direction enumeration."""

    UP = 0
    RIGHT = 1
    DOWN = 2
    LEFT = 3


class PipeStyle(IntEnum):
    """Available pipe styles."""

    HEAVY = 0
    CURVED = 1
    LIGHT = 2
    DOUBLE = 3
    KNOBBY = 4
    ANGLES = 5
    DOTS = 6
    DOTS_O = 7
    SLASHES = 8
    MIXED = 9


@dataclass
class PipeConfig:
    """Configuration for the pipes application."""

    pipes: int
    fps: int
    steady: int
    limit: int
    random_start: bool
    bold: bool
    color: bool
    keep_style: bool
    colors: list[int]
    pipe_types: list[int]


@dataclass
class Pipe:
    """Represents a single pipe with its state."""

    x: int
    y: int
    direction: Direction
    pipe_type: int
    color: int
    attr: int
